#include <stdio.h>
#include <string.h>
#include <stdlib.h>

typedef unsigned char byte;

/* Is this bank empty? */
int IsEmptyBank(const byte *buf_in, const int size_in, const int bank) {
	int i;
	
	for(i = (bank * 0x8000) + 0x200; i < ((bank + 1) * 0x8000) + 0x200; i++) {
		if(i >= size_in) return -1;		/* out of buffer! */
		if(buf_in[i] != 0) return 0;	/* not empty */
	}
	
	return 1;	/* This bank is empty! */
}

/* Search empty bank */
int SearchEmptyBank(const byte *buf_in, const int size_in) {
	int result, bank;
	bank = 0;
	
	while(1) {
		result = IsEmptyBank(buf_in, size_in, bank);
		
		if(result == -1) return -1;		/* no empty bank... */
		if(result == 1) break;			/* found an empty bank! */
		
		bank++;							/* try next bank */
	}
	
	return bank;
}

/* Write Pipes.bin to rom */
void WritePipeCode(byte *rom_buf, const int rom_size, const byte *pipe_buf, const int pipe_size, const int bank) {
	int i, offset;
	
	for(i = 0; i < pipe_size; i++) {
		offset = ((bank * 0x8000) + 0x200) + i;
		if(offset >= rom_size) break;	/* out of buffer! */
		rom_buf[offset] = pipe_buf[i];
	}
}

/* Write pipe code's bank number */
void WritePipeOffset(byte *rom_buf, const int rom_size, const int bank) {
	if(0x378B7 >= rom_size) return;	/* out of buffer! */
	
	/* No custom blocks inserted yet */
	if((rom_buf[0x37890] == 0x4C) && (rom_buf[0x37891] == 0x02) && 
	   (rom_buf[0x37892] == 0xF6) && (rom_buf[0x37893] == 0xEA))
	{
		rom_buf[0x37890] = 0xEA;
		rom_buf[0x37891] = 0xEA;
		rom_buf[0x37892] = 0xEA;
		rom_buf[0x37893] = 0xEA;
		
		rom_buf[0x378A0] = 0xEA;
		rom_buf[0x378A1] = 0xEA;
		rom_buf[0x378A2] = 0xEA;
		rom_buf[0x378A3] = 0xEA;
		
		rom_buf[0x378B0] = 0xEA;
		rom_buf[0x378B1] = 0xEA;
		rom_buf[0x378B2] = 0xEA;
		rom_buf[0x378B3] = 0xEA;
	}
	
	/* 0x37894:22 40 80 xx */
	rom_buf[0x37894] = 0x22;
	rom_buf[0x37895] = 0x40;
	rom_buf[0x37896] = 0x80;
	rom_buf[0x37897] = bank;
	
	/* 0x378A4:22 80 80 xx */
	rom_buf[0x378A4] = 0x22;
	rom_buf[0x378A5] = 0x80;
	rom_buf[0x378A6] = 0x80;
	rom_buf[0x378A7] = bank;
	
	/* 0x378B4:22 C0 80 xx */
	rom_buf[0x378B4] = 0x22;
	rom_buf[0x378B5] = 0xC0;
	rom_buf[0x378B6] = 0x80;
	rom_buf[0x378B7] = bank;
}

/* Main function */
int main(int argc, char *argv[]) {
	FILE *rom_fp;
	byte *rom_buf;
	int rom_size;
	FILE *pipe_fp;
	byte *pipe_buf;
	int pipe_size;
	int bank;
	
	/* Load Pipes.bin */
	pipe_fp = fopen("Pipes.bin", "rb");
	if(!pipe_fp) {
		printf("Failed to load Pipes.bin\n");
		return -1;
	}
	
	fseek(pipe_fp, 0, SEEK_END);
	pipe_size = ftell(pipe_fp);
	fseek(pipe_fp, 0, SEEK_SET);
	
	pipe_buf = (byte*)malloc(pipe_size);
	fread(pipe_buf, pipe_size, 1, pipe_fp);
	fclose(pipe_fp);
	
	/* Show Usage */
	if(argc <= 1) {
		printf("pipeinstall version0.01\n");
		printf("Created by purplebridge001\n");
		printf("Usage: %s [ROM Filename]\n", argv[0]);
		free(pipe_buf);
		return -2;
	}
	
	/* Open ROM */
	rom_fp = fopen(argv[1], "rb+");
	if(!rom_fp) {
		printf("Failed to load %s\n", argv[1]);
		free(pipe_buf);
		return -3;
	}
	
	fseek(rom_fp, 0, SEEK_END);
	rom_size = ftell(rom_fp);
	fseek(rom_fp, 0, SEEK_SET);
	
	rom_buf = (byte*)malloc(rom_size);
	fread(rom_buf, rom_size, 1, rom_fp);
	
	/* Search free space */
	bank = SearchEmptyBank(rom_buf, rom_size);
	
	if(bank == -1) {
		printf("No free space found in this rom\n");
		free(rom_buf);
		fclose(rom_fp);
		free(pipe_buf);
		return -4;
	}
	
	/* Write Pipes.bin to the rom */
	WritePipeCode(rom_buf, rom_size, pipe_buf, pipe_size, bank);
	WritePipeOffset(rom_buf, rom_size, bank);
	
	/* Save */
	fseek(rom_fp, 0, SEEK_SET);
	fwrite(rom_buf, rom_size, 1, rom_fp);
	fclose(rom_fp);
	
	printf("Installed pipe codes into bank %02X (0x%X)\n", bank, (bank * 0x8000) + 0x200);
	
	free(rom_buf);
	free(pipe_buf);
	return bank;
}
